#include "PelTools.h"

ClassImp(PelTools)

PelTools::PelTools():TNamed("gPelTools","Pellectron Tools")
{
  int i;
  gROOT->Add(this);
  mCutF = NULL;
  mCut = NULL;
  mList = NULL;
}
PelTools::~PelTools()
{
}
TList* PelTools::makeList(char* name)
{
  int i,n;
  if(mList) {delete mList; mList = NULL;}
  mList = new TList();
  if(gPad)
  {
    TList *l = gPad->GetListOfPrimitives();
    if(l)
    {
      n = l->GetSize();
      if(n>0) for(i=0;i<n;i++)
      {
        TObject *o = l->At(i);
        TString n = o->ClassName();
        if(n.BeginsWith(name)) mList->AddLast(o);
      }
    }
  }
  n = mList->GetSize();
  if(n==0) { delete mList; mList = NULL;}
  return mList;
}  

void PelTools::del(char* name)
{
  if(gROOT->FindObjectAny(name)) delete gROOT->FindObjectAny(name);
  return;
}
void PelTools::openFile(char *name, char* flag)
{
  if(mCutF) closeFile(); 
  mCutF = new TFile(name,flag);
}
void PelTools::createFile(char* name)
{
  openFile(name,"RECREATE");
}
void PelTools::createFileGUI()
{
  TGFileInfo fi;
  new TGFileDialog(gClient->GetRoot(),gClient->GetRoot(),kFDSave,&fi);
  if(fi.fFilename) openFile(fi.fFilename,"RECREATE");
}
void PelTools::openFileGUI()
{
  TGFileInfo fi;
  new TGFileDialog(gClient->GetRoot(),gClient->GetRoot(),kFDOpen,&fi);
  if(fi.fFilename)openFile(fi.fFilename,"UPDATE");
}
void PelTools::closeFile()
{
  if(mCutF) {mCutF->Close(); delete mCutF; mCutF=NULL;}
}
void PelTools::listFile()
{
  if(!mCutF) { message(1,"Warning","There is no cut file open"); return;}
  mCutF->ls();
}
void PelTools::saveCut(char* name)
{
  if(!mCutF) { message(1,"Warning","There is no cut file open"); return;}
  TCutG *cc = getCurrCut();
  if(!cc) { message(1,"Warning","There is no cut available for saving"); return;}
  cc->SetName(name);
  mCutF->cd();
  cc->Write();
}
TCutG* PelTools::getCurrCut()
{
  TCutG *cc = NULL;
  if(!cc) cc = mCut;
  if(!cc) cc = (TCutG*)gROOT->FindObjectAny("CUTG");
  mCut = cc;
  return cc;
}
void PelTools::saveCutGUI()
{
  if(!mCutF) { message(1,"Warning","There is no cut file open"); return;}
  TCutG *cc = getCurrCut();
  if(!cc) { message(1,"Warning","There is no cut available for saving"); return;}
  
  char *title = " Save Cut";
  TString label = "   Enter the cut identifier   ";
  TString output = " ";
  int stat;
  new TDialog(title,1,&label,&output,&stat);
  char* name = (char*)output.Data();
  if(stat==1) saveCut(name);
}
TCutG* PelTools::getCut(char* name)
{
  if(!mCutF) { message(1,"Warning","There is no cut file open"); return NULL;}
  mCut = (TCutG*)mCutF->Get(name);
  return mCut;
}
TCutG* PelTools::getCutGUI()
{
  if(!mCutF) { message(1,"Warning","There is no cut file open"); return NULL;}  
  char *title = " Load Cut";
  TString label = "   Enter the cut identifier   ";
  TString output = " ";
  int stat;
  new TDialog(title,1,&label,&output,&stat);
  char* name = (char*)output.Data();  
  if(stat==1) mCut = getCut(name);
  if(mCut) mCut->Draw();
  return mCut;
}

TH2F* PelTools::extract(TH2F* h, TCutG* c)
{
  int i,j,nx,ny,bin;
  float X,Y,V;
  if(!h) return NULL;
  if(!c) return NULL;
  TAxis *x = h->GetXaxis();
  TAxis *y = h->GetYaxis();
  nx = x->GetNbins();
  ny = y->GetNbins();
  float xmin = x->GetBinLowEdge(1);
  float xmax = x->GetBinUpEdge(nx);
  float ymin = y->GetBinLowEdge(1);
  float ymax = y->GetBinUpEdge(ny);
  TString name = h->GetName();
  name+="_Cut";
  del((char*)name.Data());
  TString title = h->GetTitle();
  title+=" (Cut)";
  TH2F *hnew = new TH2F(name.Data(),title.Data(),nx,xmin,xmax,ny,ymin,ymax);
  for(i=1; i<=nx;i++)
    for(j=1;j<=ny;j++)
    {
      bin = h->GetBin(i,j);
      X = x->GetBinCenter(i);
      Y = y->GetBinCenter(j);
      V = h->GetBinContent(bin);
      if(c->IsInside(X,Y)) hnew->Fill(X,Y,V);
    }
  hnew->SetName(name.Data());
  gROOT->GetList()->RecursiveRemove(hnew);
  return hnew;
}

void PelTools::drawProjectionCut(int m)
{
  int i,id,n;
  TH2  *hh = NULL;
  TH2F *mHC = NULL;
  TH1D *mP = NULL;
  TCutG *cc = getCurrCut();
  if(!cc) { message(1,"Warning","There is no cut available in the memory"); return;}
  TList *l = makeList("TH2");
  if(l) 
  {
    n = l->GetSize();
    for(i=0;i<n;i++)
    {
      hh = (TH2*)l->At(i);
      if(hh)
      {
        mHC = extract((TH2F*)hh,cc);
        if(mHC)
        {
          TString n = mHC->GetName();
          if(m==0) n+="_px";
          else n+="_py";
          del((char*)n.Data());
          if(m==0) mP = mHC->ProjectionX();
          else mP = mHC->ProjectionY();
          mP->SetName(n.Data());
          mP->Draw();
          //gROOT->GetList()->RecursiveRemove(mP);
        }
      }
    }
  } else { message(1,"Warning","There is no 2D histogram in the canvas"); return;}
  return;
  return;
}
void PelTools::drawTH2Cut()
{
  int i,id,n;
  TH2  *hh = NULL;
  TH2F *mHC = NULL;
  TCutG *cc = getCurrCut();
  if(!cc) { message(1,"Warning","There is no cut available in the memory"); return;}
  TList *l = makeList("TH2");
  if(l) 
  {
    n = l->GetSize();
    for(i=0;i<n;i++)
    {
      hh = (TH2*)l->At(i);
      if(hh)
      {
        mHC = extract((TH2F*)hh,cc);
        if(mHC) mHC->Draw();
      }
    }
  } else { message(1,"Warning","There is no 2D histogram in the canvas"); return;}
  return;
}
void PelTools::fitGUI()
{
  int i,n=0;
  TPeakFitGUI *gui = NULL;
  if(!gPad) { message(1,"Warning","There is no TCanvas open"); return;}
  TList *l = makeList("TH1");
  TH1 *hh=NULL;
  if(l)
  {
    n = l->GetSize();
    //l->ls();
    for(i=0;i<n;i++)
    {
      hh = (TH1*)l->At(i);
      if(hh) gui = new TPeakFitGUI(hh,gPad);
    }
  } else message(1,"Warning","There is no 1D histogram displayed");   
  return;
}
float PelTools::counts(TH1D* h,float x1,float x2)
{
  int bin1 = h->FindBin(x1);
  int bin2 = h->FindBin(x2);
  float integral = h->Integral(bin1,bin2-1);
  return integral;
}
float PelTools::counts(TH2D* h,float x1,float x2,float y1, float y2)
{
  int bin1 = h->GetXaxis()->FindBin(x1);
  int bin2 = h->GetXaxis()->FindBin(x2);
  int bin3 = h->GetYaxis()->FindBin(y1);
  int bin4 = h->GetYaxis()->FindBin(y2);
  float integral = h->Integral(bin1,bin2-1,bin3,bin4-1);
  return integral;
}
float PelTools::integral(TH1D* h,float x1,float x2)
{
  int bin1 = h->FindBin(x1);
  int bin2 = h->FindBin(x2);
  float integral = h->Integral(bin1,bin2-1,"width");
  return integral;
}
float PelTools::integral(TH2D* h,float x1,float x2,float y1, float y2)
{
  int bin1 = h->GetXaxis()->FindBin(x1);
  int bin2 = h->GetXaxis()->FindBin(x2);
  int bin3 = h->GetYaxis()->FindBin(y1);
  int bin4 = h->GetYaxis()->FindBin(y2);
  float integral = h->Integral(bin1,bin2-1,bin3,bin4-1,"width");
  return integral;
}
void PelTools::integralGUI()
{
  int i,id,n;
  float xmin,xmax,ymin,ymax;
  TList *l = NULL;
  bool noHist = true;
  char line[200];
  if(!gPad) { message(1,"Warning","There is no TCanvas open"); return;}
  TH1D *h1=NULL;
  TH2D *h2=NULL;
  xmin = gPad->GetUxmin();
  xmax = gPad->GetUxmax();
  ymin = gPad->GetUymin();
  ymax = gPad->GetUymax();
  
  l = makeList("TH1");
  if(l)
  {
    noHist = false;
    n = l->GetSize();
    for(i=0;i<n;i++) 
    {
      h1 = (TH1D*) l->At(i);
      if(h1) 
      {
        log(10,"%s 1D integral in the range:",h1->GetName());
        log(10,"   x = (%e, %e)",xmin,xmax);
        log(10,"   Counts   = %e",counts(h1,xmin,xmax));
        log(10,"   Integral = %e  (multiplied by bin size)",integral(h1,xmin,xmax));
      }
    }
  }
  l = makeList("TH2");
  if(l)
  {
    noHist = false;
    n = l->GetSize();
    for(i=0;i<n;i++) 
    {
      h2 = (TH2D*) l->At(i);
      if(h2)
      {
        log(10,"%s 2D integral in the range:",h1->GetName());
        log(10,"   x = (%e, %e)",xmin,xmax);
        log(10,"   y = (%e, %e)",ymin,ymax);
        log(10,"   Counts   = %e",counts(h2,xmin,xmax,ymin,ymax));
        log(10,"   Integral = %e  (multiplied by bin size)",integral(h2,xmin,xmax,ymin,ymax));
      }
    }
  }
  if(noHist) message(1,"Warning","There is no histogram in the current Canvas/pad");
  return ;
}
void PelTools::integralCutGUI()
{
  int i,id,n;
  TList *l = NULL;
  char line[200];
  if(!gPad) { message(1,"Warning","There is no TCanvas open"); return;}
  TH2F *h2=NULL;
  TH2F *h3=NULL;
  TCutG *cc = getCurrCut();
  if(!cc) { message(1,"Warning","There is no cut available in the memory"); return;}
  l = makeList("TH2");
  if(l)
  {
    n = l->GetSize();
    for(i=0;i<n;i++) 
    {
      h2 = (TH2F*) l->At(i);
      if(h2)
      {
        h3 = extract(h2,cc);
        log(10,"%s integral inside the graphical cut = %e",h2->GetName(),h3->Integral());
        delete h3;
      }
    }
  }
  else message(1,"Warning","There is no 2D histogram in the current Canvas/pad");
  return ;
}
void PelTools::exportAsciiGUI()
{  
  int i,n;
  if(!yesno("Export .DAT","This function will export  \nall histograms in the current Canvas/pad  \nto an ASCII file  \n\nWould you like to continue?  ")) return;
  if(!gPad) { message(1,"Warning","No active Canvas/pad"); return;}
  TList *l = makeList("TH1");
  if(!l) { message(1,"Warning","There is no 1D histograms in the current Canvas/pad"); return;}
  n = l->GetSize();
  float xmin = gPad->GetUxmin();
  float xmax = gPad->GetUxmax();
  for(i=0;i<n;i++)
  {
    TH1 *h = (TH1*)l->At(i);
    if(h)
    {
      TString name = h->GetName();
      name+=".dat";
      exportAscii(h,(char*)name.Data(),xmin,xmax);
    }
  }
}
void PelTools::exportAscii(TH1* h, char* file, float xmin, float xmax)
{  
  int j;
  float x,y,e;
  if(h)
  {
    int ibin_min = h->FindBin(xmin);
    int ibin_max = h->FindBin(xmax);
    ofstream out(file);
    char line[100];
    for(j=ibin_min; j<=ibin_max;j++)
    {
      x = h->GetBinCenter(j);
      y = h->GetBinContent(j);
      e = h->GetBinError(j);
      sprintf(line,"%6.4e     %6.4e     %6.4e",x,y,e);
      out << line<<endl;
    }
    out.close();
    log(10,"%s is exported to file << %s >> ",h->GetName(),file);
    log(10,"   limits in x = (%e, %e)",xmin,xmax);
  }
}
void PelTools::importAsciiGUI()
{  
  TGFileInfo fi;
  if(!yesno("Import .DAT","This function will import  \nthe data in an ASCII file and will  \ncreate an histogram.  \n\nWould you like to continue?  ")) return;
  new TGFileDialog(gClient->GetRoot(),gClient->GetRoot(),kFDOpen,&fi);
  if(fi.fFilename)
  {
    TDirectory *dir = (TDirectory*)gROOT->Get("Scan_Histograms");
    if(!dir) dir = (TDirectory*)gROOT->Get("SPM_Histograms");
    if(dir) dir->cd();
    else gROOT->cd();
    importAscii(fi.fFilename);
  }
}
TH1F* PelTools::importAscii(char* file)
{
  int np = 0;
  int i,j;
  float x[4096],y[4096],e[4096];
  char line[100];
  TH1F* hist = NULL;
  TString token=";, \t";
  ifstream input(file);
  log(10,"Importing file << %s >> ",file);
  while(!input.eof())
  {
    input.getline(line,100);
    TString L = line;
    TObjArray *o = L.Tokenize(token);
    int ni = 0;
    int no = o->GetSize();
    for(i=0;i<no;i++) if(!o->At(i)) {ni = i; break;}
    if(ni>0)
    {
      if(ni<2) log(10,"  *** Line %d has less than 2 columns. Skipping...",np+1);
      else
      {
        if(ni>3) log(10,"  *** Line %d has more than 3 columns. Will discard extra columns.",np+1);
	
	TString item = ((TObjString*)o->At(0))->GetString();
	x[np] = atof(item.Data());
	
	item = ((TObjString*)o->At(1))->GetString();
	y[np] = atof(item.Data());
	
	if(ni>=3)
	{
	  item = ((TObjString*)o->At(2))->GetString();
	  e[np] = atof(item.Data());
	}
	else e[np] = 0;
	np++;
      }
    }
    delete o; 
  }
  log(10,"  *** read %d lines from file. Creating histogram...",np);
  x[np] = x[np-1]+0.01;
  hist = new TH1F(file,file,np,x);
  for(i = 0;i<np;i++)
  {
    hist->SetBinContent(i+1,y[i]);
    hist->SetBinError(i+1,e[i]);
  }
  return hist;
}
void PelTools::createCut()
{  
  mCut = NULL;
  gROOT->SetEditorMode("CutG");
}
void PelTools::menu()
{  
  TList *l = gROOT->GetList();
  int id=-1;
  char line[200];
  char command[400];
  int n = l->GetSize();
  for(int i=0;i<n;i++)
  {
    if(!strcmp(l->At(i)->GetName(),"gPelTools")) { id=i; break;}
  }
  if(id==-1) return;
  sprintf(line,"((PelTools*)gROOT->GetList()->At(%d))",id);
  TControlBar *ControlBar = new TControlBar("vertical", "PelTools");
  
  ControlBar->AddButton("browser","{b = new TBrowser();}", "Start the ROOT Browser");
  ControlBar->AddButton("canvas","{c = new TCanvas();}",  "Start the ROOT Canvas");
  
  ControlBar->AddSeparator();
  
  sprintf(command,"%s->createFileGUI();",line);
  ControlBar->AddButton("new cut file",command,"Create a new banana file"); 
   
  sprintf(command,"%s->openFileGUI();",line);
  ControlBar->AddButton("open cut file",command,"Open a banana file");
  
  sprintf(command,"%s->closeFile();",line);
  ControlBar->AddButton("close cut file",command,"Close a banana file");  
  
  sprintf(command,"%s->listFile();",line);
  ControlBar->AddButton("list cut file",command,"List contentes of a banana file");
  
  sprintf(command,"%s->exportAsciiGUI();",line);
  ControlBar->AddButton("export .DAT",command,"Exports histograms in current Pad to ASCII files");
  
  sprintf(command,"%s->importAsciiGUI();",line);
  ControlBar->AddButton("import .DAT",command,"Imports ASCII file (.DAT) and create histogram");
  
  ControlBar->AddSeparator();
  
  sprintf(command,"%s->createCut();",line);
  ControlBar->AddButton("create cut",command,"Create graphical cut");  
  
  sprintf(command,"%s->saveCutGUI();",line);
  ControlBar->AddButton("   save current cut   ",command,"Save current graphical mCut into file"); 
   
  sprintf(command,"%s->getCutGUI();",line);
  ControlBar->AddButton("load cut",command,"Load graphical mCut from file");
  
  sprintf(command,"if(%s->getCurrCut()) %s->getCurrCut()->Print();",line,line);  
  ControlBar->AddButton("print cut",command,"Print mCut coordinates");
  
  sprintf(command,"if(%s->getCurrCut()) %s->getCurrCut()->Draw();",line,line);  
  ControlBar->AddButton("draw cut",command,"Draw mCut coordinates");
  
  ControlBar->AddSeparator();
  
  sprintf(command,"%s->drawTH2Cut();",line);
  ControlBar->AddButton("draw TH2(Cut)",command,"Draw TH2 Histogram gated on TCutG");
  
  sprintf(command,"%s->drawProjectionCut(0);",line,line);
  ControlBar->AddButton("draw PX(Cut)",command,"Draw Projection X");
  
  sprintf(command,"%s->drawProjectionCut(1);",line,line);
  ControlBar->AddButton("draw PY(Cut)",command,"Draw Projection Y");
  
  ControlBar->AddSeparator();
  
  sprintf(command,"%s->fitGUI();",line);
  ControlBar->AddButton("fit 1D peaks ",command,"Fit 1D peaks");
  
  sprintf(command,"%s->integralGUI();",line);
  ControlBar->AddButton("integral",command,"Integral");
  
  sprintf(command,"%s->integralCutGUI();",line);
  ControlBar->AddButton("integral TH2(Cut)",command,"Integral");
    
  ControlBar->Show();
  gROOT->SaveContext();
}
