/*!\class PelTools
\author Alexandre A. P. Suaide 

<p>
This class provides the basic functionality for data analysis using ROOT. It has tools
such as projections, fiting, polygonals, etc. 
*/
#ifndef PelTools_HH
#define PelTools_HH
#include "TNamed.h"

#include "TFile.h"
#include "TControlBar.h"
#include "TCutG.h"
#include "TGLabel.h"
#include "TGTextBuffer.h"
#include "TGTextEntry.h"
#include "TGButton.h"
#include "TString.h"
#include "TH1.h"
#include "TH2.h"
#include "TF1.h"
#include "TDialog.h"
#include "TList.h"
#include "TPeakFitGUI.h"
#include <fstream>
#include "TGTextView.h"
#include "TROOT.h"
#include "TGClient.h"
#include "TGMsgBox.h"
#include "TGFileDialog.h" 
#include "TPad.h"
#include "TSpectrum.h"
#include "log.hh"
#include "TObjString.h"
#include <iostream>
using namespace std;

class PelTools : public TNamed
{
  private:    
    TFile            *mCutF;
    TList            *mList;
    TCutG            *mCut;
          
    void             del(char*);
    // create list of histograms in pad
    TList*           makeList(char*);  
        
  public:
                     PelTools();
    virtual         ~PelTools();
    
    // these methods are to be used only from the GUI
    void             createFileGUI();
    void             openFileGUI();
    void             saveCutGUI();
    void             drawTH2Cut();
    TCutG*           getCutGUI();
    void             fitGUI();
    void             integralGUI();
    void             integralCutGUI();
    void             drawProjectionCut(int);
    void             exportAsciiGUI();
    void             importAsciiGUI();

    // these are iteractive methods
    void             openFile(char *name, char* flag="");
    void             createFile(char* name);
    void             closeFile();
    void             listFile();
    void             saveCut(char* name="CUTG");
    TCutG*           getCut(char* name="CUTG");
    TCutG*           getCurrCut();
    TH2F*            extract(TH2F* h, TCutG* c);
    float            integral(TH1D*,float,float);
    float            integral(TH2D*,float,float,float,float);
    float            counts(TH1D*,float,float);
    float            counts(TH2D*,float,float,float,float);
    void             createCut();
    void             exportAscii(TH1*,char*,float,float);
    TH1F*            importAscii(char*);
    
    void             menu();
   
  ClassDef(PelTools,1)
};

#endif
