#include "adac+.h"
#include "TROOT.h"
TString baseName;
TString title;
int KSYS=0,KAUX=0,NAUX=0,NSHD=0;
int ADINIT = 0;

TList   code[5];

int main(int argc, char** argv)
{
  if(argc==1)
  {
    cout <<"usage: adac+ filename\n";
    return 0;
  }
  for(int i=0;i<5;i++) code[i].Delete();
  title = "NO TITLE";
  baseName = argv[1];
  if(baseName.EndsWith(".evs")) baseName = baseName(0,baseName.Length()-4);
  readFile((char*)baseName.Data());
  pass0();
  pass1();
  pass2();
  pass3();
  pass4();
  saveOut((char*)baseName.Data());
  dump((char*)baseName.Data());
  for(int i=0;i<5;i++) code[i].Delete();
  cout <<"NO ERRORS !!\n";
}
void dump(char* fileName)
{
  TList* CODE = &code[3];
  TString o = fileName;
  o+=".out"; 
  ofstream out(o.Data());
  out <<"TITLE = "<<title.Data()<<endl;
  out <<"KSYS = "<<KSYS<<"  KAUX = "<<KAUX<<"  NAUX = "<<NAUX<<"  NSHD = "<<NSHD<<endl;
  out <<"------------------------------------------------------\n";
  for(int i=0;i<CODE->GetSize();i++)
  {
    Program *p=(Program*) CODE->At(i);
    char U[30];
    sprintf (U,"%4d | %1d | %4d | ",p->getLine(),p->getFlag(),p->getPosition());
    out <<U;
    for(int k = 0;k<3;k++)
    {
      if(k<p->getNComp())
      {
        char line[10];
        sprintf(line,"0x%08x",p->getComp(k));
        out  <<line<<" | ";
      } 
      else 
      { 
        out <<"           | "; 
      }
    }
    out <<p->getCode()<<endl;
  }
}
void err(int C,int line)
{
  TString M;
  if(C==IOERR)             M = "I/O error";
  if(C==SYNTAXERR)         M = "Syntax Error";
  if(C==OPERERR)           M = "Operator error";
  if(C==LOOPERR)           M = "Loop error";
  if(C==LOOPNOTSTARTED)    M = "ENDLOOP without LOOP";
  if(C==NESTEDLOOPS)       M = "Nested loops";
  if(C==LOOPSIZEERR)       M = "Loop size error";
  if(C==LOOPLABELERR)      M = "Multiple labels in the beginning of the loop";
  if(C==LOOPNOTCLOSED)     M = "Loop not closed";
  if(C==VARIABLENOTDEF)    M = "Assembly time variable not defined";
  if(C==NOTVALIDCOMMAND)   M = "Syntax Error: Command invalid";
  if(C==INVALIDNUMBERCMD)  M = "Syntax Error: Wrong number of parameters";
  cout <<"ERROR: code =  "<<C<<", line = "<<line<<endl;
  cout <<M.Data()<<endl;
  if(line>=0) 
  {
    int l0 = line-3;
    int l1 = line+2;
    if(l0<0) l0=0;
    if(l1>code[0].GetSize()) l1=code[0].GetSize();
    for(int i=l0;i<l1;i++)
    {
      Program *p = (Program*)code[0].At(i);
      if(i==line-1) cout <<"----> "<< p->getCode()<<endl;
      else cout <<"      "<< p->getCode()<<endl;
    }
  }
  exit(0);
}
void readFile(char* fileName)
{
  TString file=fileName;
  file+=".evs";
  FILE *in = fopen((char*)file.Data(),"r");
  if(in==0) err(IOERR);
  char line[400];
  int n=0;
  while(!feof(in))
  {
    fgets(line,400,in);
    n++;
    TString L = line;
    L=L(0,L.Length()-1);
    L.ToUpper(); 
    Program* p = new Program();
    p->setLine(n);
    p->setCode((char*)L.Data());
    code[0].Add(p);
  }
  fclose(in);
  return;
}

// pass 0 finds the pre compiler directives, flags the comments and
// expressions
void pass0()
{
  int DEFFLAG = CLEAN;
	for(int i=0;i<code[0].GetSize();i++)
  {
    Program *p = (Program*)code[0].At(i);
    TString L  = p->getCodeClean();
    TString L0 = p->getCode();
    if(L.Length()>0 && p->getFlag()==CLEAN)
    {
      int flag=DEFFLAG;
      int j = L.Index("=");
      int value = 0;
      if(j>0) value = ATOI((char*)(L(j+1,L.Length()).Data())); 
      if(L.BeginsWith(";"))    { flag=COMMENT; }
      if(L.BeginsWith("PGM$")) { title=L(5,L.Length()); flag=PRECOMP; DEFFLAG = CLEAN;}
      if(L.BeginsWith("COM$")) { flag=PRECOMP; DEFFLAG = COMMENT; }
      if(L.BeginsWith("CSYS")) { flag=PRECOMP; KSYS = value; if(j<0) err(SYNTAXERR,i); }
      if(L.BeginsWith("CREM")) { flag=PRECOMP; KAUX = value; if(j<0) err(SYNTAXERR,i); }
      if(L.BeginsWith("CAUX")) { flag=PRECOMP; KAUX = value; if(j<0) err(SYNTAXERR,i); }
      if(L.BeginsWith("NAUX")) { flag=PRECOMP; NAUX = value; if(j<0) err(SYNTAXERR,i); }
      if(L.BeginsWith("NSHD")) { flag=PRECOMP; NSHD = value; if(j<0) err(SYNTAXERR,i); }
      if(j>0 && flag==CLEAN)   { flag=EXPRESSION; }
      if(!L0.BeginsWith(" ") && flag==CLEAN)  { flag=JUMPLABEL; }
      p->setFlag(flag);
    } else p->setFlag(COMMENT);
  }  
}

// pass 1 opens the loops
void pass1()
{
  int nL = 0;
  int loop1 = 0;
  int loop2 = 0;
  for(int i=0;i<code[0].GetSize();i++)
  {
    Program *p = (Program*)code[0].At(i);
    TString L  = p->getCodeClean();
    TString C  = p->getCode(true);
    int     F  = p->getFlag();
    int     l  = p->getLine();
    int     i0 = C.Index(" LOOP ");
    int     i1 = C.Index(" ENDL");
    if(nL==0 && i0<0)
    {
      Program *g = new Program();
      g->setLine(l);
      g->setFlag(F);
      g->setCode((char*)C.Data());
      code[1].Add(g);
    }
    if(i1>=0 && nL==0) err(LOOPNOTSTARTED,l);
    if(i0>=0 && nL>0) err(NESTEDLOOPS,l);
    if(i0>=0) { loop1 = i; nL++;}
    if(i1>=0) 
    {
      loop2 = i;
      nL--;
      Program *p0 = (Program*)code[0].At(loop1);
      Program *p1 = (Program*)code[0].At(loop1+1);
      
      // getting the size of the loop
      TString  O  = p0->getCode(true);
      int index = O.Index(" LOOP ");
      O = O(index+6,O.Length());
      int size = ATOI((char*)O.Data());
      if(size<=0) err(LOOPSIZEERR,p0->getLine());
      
      // checking if the loop starts with a label
      // can not have two different labels in the begginign of the loop
      if(p0->getFlag()==JUMPLABEL && p1->getFlag()==JUMPLABEL) err(LOOPLABELERR,p0->getLine());
      TString LABEL="";
      if(p->getFlag()==JUMPLABEL)
      {
        TString H = p0->getCode(true);
        LABEL = H(0,H.Index(" "));
      }
      for(int j = 0;j<size;j++)
      {
        for(int k=loop1+1;k<loop2;k++)
        {
          Program *p2 = (Program*)code[0].At(k);
          TString CODE = "";
          int     FLAG = p2->getFlag();
          int     LINE = p2->getLine();
          if(j==0 && k==loop1+1) 
          {
            CODE+=LABEL;
            FLAG = p0->getFlag();
          }
          CODE+=p2->getCode(true);
          Program *g = new Program();
          g->setLine(LINE);
          g->setFlag(FLAG);
          g->setCode((char*)CODE.Data());
          code[1].Add(g);
        }
      }
    }
  }
  if(nL>0) err(LOOPNOTCLOSED,loop1+1);
  for(int i=0;i<code[1].GetSize();i++)
  {
    Program *p =(Program*)code[1].At(i);
    if(p) p->process();
  }
}

// pass 2 solves assembly time variables and make the short code
// Removes expressions. This takes more time because we use the
// ROOT cint interpreter to solve the expressions
void pass2()
{
  int k0,k1, position=1;
  for(int i=0;i<code[1].GetSize();i++)
  {
    Program *p = (Program*)code[1].At(i);
    int LINE     = p->getLine();
    TString CODE = p->getCode(true);
    TString COCL = p->getCodeClean();
    int FLAG     = p->getFlag();
    TString LAB  = p->getLabel();
    TList* L   = p->getList();
    if(FLAG==EXPRESSION)
    {
      TString a = "Int_t ";
      a+=COCL;
      a+=";";
      int e;
      gROOT->ProcessLineSync(a.Data(),&e);
      if(e) err(OPERERR,LINE);
    }
    if(FLAG==CLEAN || FLAG==JUMPLABEL)
    {
      Program *n = new Program();
      code[2].Add(n);
      n->setLine(LINE);
      n->setFlag(FLAG);
      n->setCode((char*)CODE.Data());
      n->setLabel((char*)LAB.Data());
      n->setPosition(position);
      TList *N = n->getList();
      for(int j=0;j<L->GetSize();j++)
      {
        k1=0;
        TString U = ((TObjString*)L->At(j))->String();
        if(isCommand((char*)U.Data(),k0,k1)) 
        {
          N->Add(new TObjString(U));
          position+=k1;
        }
        else if(isOperator((char*)U.Data(),k0))     N->Add(new TObjString(U));
        else if(isRegister((char*)U.Data(),k0,k1))  N->Add(new TObjString(U));
        else if(isDelimiter((char*)U.Data()))       N->Add(new TObjString(U));
        else if(isLabel((char*)U.Data(),k0))        N->Add(new TObjString(U));
        else
        {
          int u = ATOI(U.Data());
          if(u!=0) N->Add(new TObjString(U));
          else if(u==0 && U.BeginsWith("0")) N->Add(new TObjString(U));
          else
          {
            //U = p->getCodeClean((char*)U.Data());
            TGlobal *g = (TGlobal*)gROOT->GetGlobal(U.Data(),kTRUE);
            if(!g) 
            {
              cout <<U.Data()<<endl;
              cout <<"Problem finding variable "<<U.Data()<<endl;
              err(VARIABLENOTDEF,LINE);
            }
            else
            {
              int* value = (int*)g->GetAddress();
              if(!value) err(VARIABLENOTDEF,LINE);
              TString UU;
              UU+=(*value);
              N->Add(new TObjString(UU));
            }
          }
        }
      }
    }
  }
}
// pass 3 makes the bit pattern
void pass3()
{
  for(int i=0;i<code[2].GetSize();i++)
  {
    Program *p = (Program*)code[2].At(i);
    int LINE     = p->getLine();
    TString CODE = p->getCode(true);
    TString COCL = p->getCodeClean();
    int FLAG     = p->getFlag();
    int POS      = p->getPosition();
    TString LAB  = p->getLabel();
    TList* L     = p->getList();
    Program *n = new Program();
    n->setLine(LINE);
    n->setCode((char*)CODE.Data());
    n->setFlag(FLAG);
    n->setLabel((char*)LAB.Data());
    n->setPosition(POS);
    TList *N = n->getList();
    code[3].Add(n);
    
    bool bitOpen = false;
    bool next    = false;
    bool saveBit = false;
    int  VAL = 0;
    for(int j=0;j<L->GetSize();j++)
    {
      TString v = ((TObjString*)L->At(j))->String();
      if(!strcmp(v.Data(),"[")) { bitOpen = true; next = true;}
      if(!strcmp(v.Data(),"]"))
      {
        if(!bitOpen) err(SYNTAXERR,i);
        bitOpen = false;
        saveBit = true;
      } 
      if(!bitOpen && !saveBit) N->Add(new TObjString(v));   
      if(!bitOpen && saveBit)  
      {
        TString A; 
        A+= VAL; 
        N->Add(new TObjString(A)); 
        saveBit = false; 
        VAL=0;
      }
      if(bitOpen)
      {
        if(next) next = false;
        else 
        {
          if(ATOI(v.Data())<=0) err(SYNTAXERR,LINE);
          VAL+=(int)pow(2.0,ATOI(v.Data())-1);
        }
      }      
    }    
  }
}
void pass4()
{
  for(int i=0;i<code[3].GetSize();i++)
  {
    Program *p = (Program*)code[3].At(i);
    int LINE     = p->getLine();
    int FLAG     = p->getFlag();
    int POS      = p->getPosition();
    TList *l     = p->getList();
    if(l->GetSize()>0)
    {
      TString cmd = ((TObjString*)l->At(0))->String();
      int type,ncom;
      if(!isCommand((char*)cmd.Data(),type,ncom)) err(NOTVALIDCOMMAND,LINE);
      switch(type)
      {
        case C_NOP:    p->addComp(0x00000000); break;
        case C_NAF:    makeNAF(p); break;
        case C_NAFX:   makeNAFX(p); break;
        case C_FLUSH:  makeFLUSH(p); break;
        case C_SPB:    makeSPB(p); break;
        case C_BRU:    makeBRU(p); break;
        case C_BRUR:   p->addComp(0x00100003); break;
        case C_SKIP:   makeSKIP(p); break;
        case C_MOV:    makeMOV(p); break;
        case C_MERG:   makeMERG(p); break;
        case C_OUT:    makeOUT(p); break;
        case C_PAR:    makePAR(p); break;
        case C_INC:    makeINC(p); break;
        case C_DEC:    makeDEC(p); break;
        case C_AND:    makeAND(p); break;
        case C_OR:     makeOR(p); break;
        case C_NOT:    makeNOT(p); break;
        case C_RSH:    makeRSH(p); break;
        case C_LSH:    makeLSH(p); break;
        case C_MOD:    makeMOD(p); break;
        case C_MUL:    makeMUL(p); break;
        case C_DIV:    makeDIV(p); break;
        case C_DLAY:   makeDLAY(p); break;
        case C_SETC:   makeSETC(p); break;
        case C_RTRN:   p->addComp(0x00100000); break;
        case C_END:    p->addComp(0x00100001); break;
        case C_INIT:   p->addComp(0x00100002); ADINIT = p->getPosition(); break;
        
        case C_CNAF:   makeCNAF(p); break;
        case C_IF:     makeIF(p); break;
        case C_ROUT:   makeROUT(p); break;
        case C_CROUT:  makeCROUT(p); break;
        case C_ABORT:  p->addComp(0x00000001); break;
        
        default: err(NOTVALIDCOMMAND,LINE);
      }
    }
  }
}

bool isOperator(char* c,int& type)
{
  if(!strcmp(c,"NONE")) { type = O_NONE; return true;}
  if(!strcmp(c,"ANY"))  { type = O_ANY;  return true;}
  if(!strcmp(c,"GT"))   { type = O_GT;   return true;}
  if(!strcmp(c,"LT"))   { type = O_LT;   return true;}
  if(!strcmp(c,"EQ"))   { type = O_EQ;   return true;}
  if(!strcmp(c,"NE"))   { type = O_NE;   return true;}
  if(!strcmp(c,"LE"))   { type = O_LE;   return true;}
  if(!strcmp(c,"GE"))   { type = O_GE;   return true;}
  return false;
}
bool isRegister(char* c,int& type,int& index)
{
  TString C=c;
  if(C.BeginsWith("PAT"))
  {
    type = R_PAT;
    if(!strcmp("PAT",c)) {index = 0; return true;}
    TString I = C(3,C.Length());
    index = ATOI(I.Data());
    if(index<0 || index>9) { index = 0; return false;}
    else return true;
  }
  if(C.BeginsWith("EX"))
  {
    type = R_EXX;
    if(!strcmp("EX",c)) {index = 0; return true;}
    TString I = C(2,C.Length());
    index = ATOI(I.Data());
    if(index<0 || index>9) { index = 0; return false;}
    else return true;
  }
  if(C.BeginsWith("CA"))
  {
    type = R_CA;
    if(!strcmp("CA",c)) {index = 0; return true;}
    TString I = C(2,C.Length());
    index = ATOI(I.Data());
    if(index<0 || index>9) { index = 0; return false;}
    else return true;
  }
  index = 0;
  return false;
}
bool isDelimiter(char* c)
{
  if(!strcmp("(",c)) return true;
  if(!strcmp(")",c)) return true;
  if(!strcmp("[",c)) return true;
  if(!strcmp("]",c)) return true;
  return false;
}
bool isCommand(char* c, int& type, int& ncom)
{
  if(!strcmp(c,"NOP"))   { type = C_NOP;   ncom = 1; return true;}
  if(!strcmp(c,"NAF"))   { type = C_NAF;   ncom = 1; return true;}
  if(!strcmp(c,"NAFX"))  { type = C_NAFX;   ncom = 1; return true;}
  if(!strcmp(c,"FLUSH")) { type = C_FLUSH; ncom = 1; return true;}
  if(!strcmp(c,"SPB"))   { type = C_SPB;   ncom = 1; return true;}
  if(!strcmp(c,"BRU"))   { type = C_BRU;   ncom = 1; return true;}
  if(!strcmp(c,"BRUR"))  { type = C_BRUR;  ncom = 1; return true;}
  if(!strcmp(c,"SKIP"))  { type = C_SKIP;  ncom = 1; return true;}
  if(!strcmp(c,"MOV"))   { type = C_MOV;   ncom = 1; return true;}
  if(!strcmp(c,"MERG"))  { type = C_MERG;  ncom = 1; return true;}
  if(!strcmp(c,"OUT"))   { type = C_OUT;   ncom = 1; return true;}
  if(!strcmp(c,"PAR"))   { type = C_PAR;   ncom = 1; return true;}
  if(!strcmp(c,"INC"))   { type = C_INC;   ncom = 1; return true;}
  if(!strcmp(c,"DEC"))   { type = C_DEC;   ncom = 1; return true;}
  if(!strcmp(c,"AND"))   { type = C_AND;   ncom = 1; return true;}
  if(!strcmp(c,"OR"))    { type = C_OR;    ncom = 1; return true;}
  if(!strcmp(c,"NOT"))   { type = C_NOT;   ncom = 1; return true;}
  if(!strcmp(c,"RSH"))   { type = C_RSH;   ncom = 1; return true;}
  if(!strcmp(c,"LSH"))   { type = C_LSH;   ncom = 1; return true;}
  if(!strcmp(c,"MOD"))   { type = C_MOD;   ncom = 1; return true;}
  if(!strcmp(c,"MUL"))   { type = C_MUL;   ncom = 1; return true;}
  if(!strcmp(c,"DIV"))   { type = C_DIV;   ncom = 1; return true;}
  if(!strcmp(c,"DLAY"))  { type = C_DLAY;  ncom = 1; return true;}
  if(!strcmp(c,"SETC"))  { type = C_SETC;  ncom = 1; return true;}
  if(!strcmp(c,"RTRN"))  { type = C_RTRN;  ncom = 1; return true;}
  if(!strcmp(c,"END"))   { type = C_END;   ncom = 1; return true;}
  if(!strcmp(c,"INIT"))  { type = C_INIT;  ncom = 1; return true;}
  
  if(!strcmp(c,"CNAF"))  { type = C_CNAF;  ncom = 2; return true;}
  if(!strcmp(c,"IF"))    { type = C_IF;    ncom = 2; return true;}
  if(!strcmp(c,"ROUT"))  { type = C_ROUT;  ncom = 2; return true;}
  if(!strcmp(c,"CROUT")) { type = C_CROUT; ncom = 3; return true;}
  
  return false;
}
bool isLabel(char* c,int& position)
{  
  for(int i=0;i<code[3].GetSize();i++)
  {
    Program *p =(Program*)code[3].At(i);
    TString L = p->getLabel();
    if(L.Length()>0) if(!strcmp(c,(char*)L.Data())) { position = p->getPosition(); return true;}
  }
  for(int i=0;i<code[1].GetSize();i++)
  {
    Program *p =(Program*)code[1].At(i);
    TString L = p->getLabel();
    if(L.Length()>0) if(!strcmp(c,(char*)L.Data())) { position = p->getPosition(); return true;}
  }
  return false;
}
void  makeNAF(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=4 && l->GetSize()!=5) err(INVALIDNUMBERCMD,P->getLine());
  int N = ATOI((((TObjString*)l->At(1))->String()).Data());
  int A = ATOI((((TObjString*)l->At(2))->String()).Data());
  int F = ATOI((((TObjString*)l->At(3))->String()).Data());
  if(N<1 || N>25) err(PARERR,P->getLine());
  if(A<0 || A>16) err(PARERR,P->getLine());
  if(F<0 || F>26) err(PARERR,P->getLine());
  TString REG = "CA0";
  if(l->GetSize()==5) REG = ((TObjString*)l->At(4))->String();
  int rtype, r=0;
  bool isReg = isRegister((char*)REG.Data(),rtype,r);
  if(!isReg) err(PARERR,P->getLine());
  int NAF = NAFGEN(N,A,F);
  
  unsigned int CMD = 0x00800000;
  
  CMD = CMD | (rtype << 20) | (r << 16) | NAF;
  P->addComp(CMD);
}
void  makeNAFX(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=4 && l->GetSize()!=5) err(INVALIDNUMBERCMD,P->getLine());
  TString EXN = ((TObjString*)l->At(1))->String();
  TString EXA = ((TObjString*)l->At(2))->String();
  TString EXF = ((TObjString*)l->At(3))->String();
  TString REG = "CA0";
  if(l->GetSize()==5) REG = ((TObjString*)l->At(4))->String();

  int RT,rN,rA,rF;
  if(!isRegister((char*)EXN.Data(),RT,rN)) err(PARERR,P->getLine());;
  if(RT!=R_EXX) err(PARERR,P->getLine());;
  if(!isRegister((char*)EXA.Data(),RT,rA)) err(PARERR,P->getLine());;
  if(RT!=R_EXX) err(PARERR,P->getLine());;
  if(!isRegister((char*)EXF.Data(),RT,rF)) err(PARERR,P->getLine());;
  if(RT!=R_EXX) err(PARERR,P->getLine());;
  int rtype, r=0;
  bool isReg = isRegister((char*)REG.Data(),rtype,r);
  if(!isReg) err(PARERR,P->getLine());
  
  unsigned int CMD = 0x00808000;
  CMD = CMD | (rtype << 20) | (r << 16) | (rN << 8)| (rA << 4)| rF;
  P->addComp(CMD);
}
void  makeFLUSH(Program* P){}
void  makeSPB(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());
  TString L = ((TObjString*)l->At(1))->String();
  int value = 0;
  bool isL = isLabel((char*)L.Data(),value);
  if(!isL) value = ATOI(L.Data());
  if(value<=0 || value > 0xFFFF) err(PARERR,P->getLine());
	if(value>ADINIT) value-=ADINIT;
  
  P->addComp(0x00400000 | (value & 0xFFFF));
}
void  makeBRU(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());
  TString L = ((TObjString*)l->At(1))->String();
  int value = 0;
  bool isL = isLabel((char*)L.Data(),value);
  if(!isL) value = ATOI(L.Data());
  if(value<=0 || value > 0xFFFF) err(PARERR,P->getLine());
	if(value>ADINIT) value-=ADINIT;
    
  P->addComp(0x00500000 | (value & 0xFFFF));
}
void  makeSKIP(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=4) err(INVALIDNUMBERCMD,P->getLine());
  TString  REG1 = ((TObjString*)l->At(1))->String();
  TString  OPER = ((TObjString*)l->At(2))->String();
  TString  REG2 = ((TObjString*)l->At(3))->String();
  
  int otype=0, r1type=0, r2type=0, r1=0,r2=0;
  unsigned int CMD = 0x80000000;
  unsigned int V = 0;

  if(!isOperator((char*)OPER.Data(),otype)) err(OPERERR,P->getLine());
  bool isR1 = isRegister((char*)REG1.Data(),r1type,r1);
  bool isR2 = isRegister((char*)REG2.Data(),r2type,r2);
  if(!isR1) err(PARERR,P->getLine());
  if(!isR2) V = ATOI(REG2.Data());
  if(V<0 || V>0xFFFF) err(PARERR,P->getLine());
  
  CMD = CMD | (otype << 28);
  
  if(isR1) CMD = CMD | ( r1type << 26);
  else r1 = 0;
  if(isR2) CMD = CMD | ( r2type << 24);
  else r2 = 0;
  
  CMD = CMD | ( (r1 & 0xF) << 20 ) | ( (r2 & 0xF) << 16) | (V & 0xFFFF);
  P->addComp(CMD);
}
int makeClass2(Program* p, char* PB ,char* PA,bool NUMA)
{
  int CMD = 0;
  int rAtype=0,rA=0,rBtype=0,rB=0;
  bool isRA = isRegister(PA,rAtype,rA);
  bool isRB = isRegister(PB,rBtype,rB);
  if(NUMA && isRA) err(PARERR,p->getLine());
  if(!isRB) err(PARERR,p->getLine());
  
  if(isRA) CMD = CMD | (rAtype << 22);
  else rA = ATOI(PA);
  CMD = CMD | (rBtype << 20);
  CMD = CMD | (rB << 16);
  CMD = CMD | (rA & 0xFFFF);
  return CMD;
}
void  makeMERG(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=2 && l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());
  
  TString  RB = "CA";
  if(l->GetSize()==3) RB = ((TObjString*)l->At(2))->String();
  TString  RA = ((TObjString*)l->At(1))->String();
  
  unsigned int CMD = 0x01000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeMOV(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RA = ((TObjString*)l->At(1))->String();
  TString  RB = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x02000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);  
}
void  makeINC(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3 && l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = "1";
  if (l->GetSize()==3) RA=((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x03000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeDEC(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3 && l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = "1";
  if (l->GetSize()==3) RA=((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x04000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeAND(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x05000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeOR(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x06000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeNOT(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  
  unsigned int CMD = 0x07000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),"0",false);
  P->addComp(CMD);
}
void  makeRSH(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x08000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeLSH(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x09000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeMOD(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x0A000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeMUL(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x0B000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}
void  makeDIV(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=3) err(INVALIDNUMBERCMD,P->getLine());

  TString  RB = ((TObjString*)l->At(1))->String();
  TString  RA = ((TObjString*)l->At(2))->String();
  
  unsigned int CMD = 0x0C000000;
  CMD = CMD | makeClass2(P,(char*)RB.Data(),(char*)RA.Data(),false);
  P->addComp(CMD);
}

void  makePAR(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());
  
  TString C = ((TObjString*)l->At(1))->String();
  
  Program *p1 = new Program();
  TList *l1   = p1->getList();
  p1->setLine(P->getLine());
  p1->setCode(P->getCode(true));  
  p1->setPosition(P->getPosition());  
  l1->Add(new TObjString("OUT"));
  l1->Add(new TObjString(C));
  makeOUT(p1);
  P->addComp(p1->getComp(0) | 0x00080000);
  delete p1;
}
void  makeOUT(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());
  TString REG = ((TObjString*)l->At(1))->String();
  int rtype=0,r=0, value = 0;
  bool isR = isRegister((char*)REG.Data(),rtype,r);
  if(!isR) value = ATOI((char*)REG.Data());
  
  unsigned int CMD = 0x00F00000;
  CMD = CMD | (rtype << 16) | r | value;
  P->addComp(CMD);  
}

void  makeDLAY(Program* P)
{  
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());
  TString L = ((TObjString*)l->At(1))->String();
  int value = ATOI(L.Data());
  if(value<=0 || value > 0xFFFF) err(PARERR,P->getLine());
  
  P->addComp(0x00700000 | (value & 0xFFFF));
}
void  makeSETC(Program* P)
{  
  TList *l = P->getList();
  if(l->GetSize()!=2) err(INVALIDNUMBERCMD,P->getLine());
  TString L = ((TObjString*)l->At(1))->String();
  int value = ATOI(L.Data());
  if(isLabel((char*)L.Data(),value)) err(PARERR,P->getLine());
  if(value<=0 || value > 0xF) err(PARERR,P->getLine());
  
  P->addComp(0x00600000 | (value && 0xFFFF));
}
void  makeCNAF(Program* P)
{  
  TList *l = P->getList();
  if(l->GetSize()!=5 && l->GetSize()!=6) err(INVALIDNUMBERCMD,P->getLine());
  TString REG = "CA0";
  TString  C = ((TObjString*)l->At(1))->String(); 
  TString  N = ((TObjString*)l->At(2))->String();
  TString  A = ((TObjString*)l->At(3))->String();
  TString  F = ((TObjString*)l->At(4))->String();
  if(l->GetSize()==6) REG = ((TObjString*)l->At(5))->String();
  
  Program *p1 = new Program();
  TList *l1   = p1->getList();
  p1->setLine(P->getLine());
  p1->setCode(P->getCode());  
  p1->setPosition(P->getPosition());  
  l1->Add(new TObjString("SETC"));
  l1->Add(new TObjString(C));
  makeSETC(p1);
  P->addComp(p1->getComp(0));
  delete p1;
  
  Program *p2 = new Program();
  TList *l2   = p2->getList();
  p2->setLine(P->getLine());
  p2->setCode(P->getCode());  
  p2->setPosition(P->getPosition());  
  l2->Add(new TObjString("NAF"));
  l2->Add(new TObjString(N));
  l2->Add(new TObjString(A));
  l2->Add(new TObjString(F));
  l2->Add(new TObjString(REG));
  makeNAF(p2);
  P->addComp(p2->getComp(0));
  delete p2;    
}
void  makeIF(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=5) err(INVALIDNUMBERCMD,P->getLine());
  TString  R1 = ((TObjString*)l->At(1))->String();
  TString  OP = ((TObjString*)l->At(2))->String();
  TString  R2 = ((TObjString*)l->At(3))->String();
  TString  LA = ((TObjString*)l->At(4))->String();
  
  int otype = 0;
  if(!isOperator((char*)OP.Data(),otype)) err(OPERERR,P->getLine());
  if(otype==O_ANY)  OP = "NONE";
  if(otype==O_NONE) OP = "ANY";
  if(otype==O_GT)   OP = "LE";
  if(otype==O_LT)   OP = "GE";
  if(otype==O_GE)   OP = "LT";
  if(otype==O_LE)   OP = "GT";
  if(otype==O_NE)   OP = "EQ";
  if(otype==O_EQ)   OP = "NE";  
  
  Program *p1 = new Program();
  TList *l1   = p1->getList();
  p1->setLine(P->getLine());
  p1->setCode(P->getCode());  
  p1->setPosition(P->getPosition());  
  l1->Add(new TObjString("SKIP"));
  l1->Add(new TObjString(R1));
  l1->Add(new TObjString(OP));
  l1->Add(new TObjString(R2));
  makeSKIP(p1);
  P->addComp(p1->getComp(0));
  delete p1;
  
  Program *p2 = new Program();
  TList *l2   = p2->getList();
  p2->setLine(P->getLine());
  p2->setCode(P->getCode());  
  p2->setPosition(P->getPosition());  
  l2->Add(new TObjString("BRU"));
  l2->Add(new TObjString(LA));
  makeBRU(p2);
  P->addComp(p2->getComp(0));
  delete p2;
}
void  makeROUT(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=4 && l->GetSize()!=5) err(INVALIDNUMBERCMD,P->getLine());
  TString REG = "CA0";
  TString  N = ((TObjString*)l->At(1))->String();
  TString  A = ((TObjString*)l->At(2))->String();
  TString  F = ((TObjString*)l->At(3))->String();
  if(l->GetSize()==5) REG = ((TObjString*)l->At(4))->String();
  
  Program *p1 = new Program();
  TList *l1   = p1->getList();
  p1->setLine(P->getLine());
  p1->setCode(P->getCode());  
  p1->setPosition(P->getPosition());  
  l1->Add(new TObjString("NAF"));
  l1->Add(new TObjString(N));
  l1->Add(new TObjString(A));
  l1->Add(new TObjString(F));
  l1->Add(new TObjString(REG));
  makeNAF(p1);
  P->addComp(p1->getComp(0));
  delete p1;
  
  Program *p2 = new Program();
  TList *l2   = p2->getList();
  p2->setLine(P->getLine());
  p2->setCode(P->getCode());  
  p2->setPosition(P->getPosition());  
  l2->Add(new TObjString("OUT"));
  l2->Add(new TObjString(REG));
  makeOUT(p2);
  P->addComp(p2->getComp(0));
  delete p2;
}
void  makeCROUT(Program* P)
{
  TList *l = P->getList();
  if(l->GetSize()!=5 && l->GetSize()!=6) err(INVALIDNUMBERCMD,P->getLine());
  TString REG = "CA0";
  TString  C = ((TObjString*)l->At(1))->String(); 
  TString  N = ((TObjString*)l->At(2))->String();
  TString  A = ((TObjString*)l->At(3))->String();
  TString  F = ((TObjString*)l->At(4))->String();
  if(l->GetSize()==6) REG = ((TObjString*)l->At(5))->String();
  
  Program *p1 = new Program();
  TList *l1   = p1->getList();
  p1->setLine(P->getLine());
  p1->setCode(P->getCode());  
  p1->setPosition(P->getPosition());  
  l1->Add(new TObjString("SETC"));
  l1->Add(new TObjString(C));
  makeSETC(p1);
  P->addComp(p1->getComp(0));
  delete p1;
  
  Program *p2 = new Program();
  TList *l2   = p2->getList();
  p2->setLine(P->getLine());
  p2->setCode(P->getCode());  
  p2->setPosition(P->getPosition());  
  l2->Add(new TObjString("ROUT"));
  l2->Add(new TObjString(N));
  l2->Add(new TObjString(A));
  l2->Add(new TObjString(F));
  l2->Add(new TObjString(REG));
  makeROUT(p2);
  P->addComp(p2->getComp(0));
  P->addComp(p2->getComp(1));
  delete p2;    
}

int ATOI(char* v)
{
  TString N = v;
  if( N.BeginsWith("0X")  ||  N.EndsWith("H")) return strtol(v,NULL,16);
  if(!N.BeginsWith("0X")  &&  N.EndsWith("B")) return strtol(v,NULL,2);
  if( N.BeginsWith("0B")  && !N.EndsWith("H")) 
  {
    N=N(2,N.Length());
    return strtol(N.Data(),NULL,2);
  }
  return atoi(v);
}
int ATOI(const char* v)
{
  return ATOI((char*)v);
}
void saveOut(char * fileName)
{
  TString file=fileName;
  file+=".evp";
  FILE *ou = fopen((char*)file.Data(),"w");
  if(ou==0) err(IOERR);
  TList* CODE = &code[3];  
  int NC = 0;
  for(int i=0;i<CODE->GetSize();i++)
  {
    Program *p=(Program*) CODE->At(i);
    NC+=p->getNComp();
  }
  fprintf(ou,"%s\n",title.Data());
  fprintf(ou,"%d %d  %d  %d  %d %d\n",NC,KSYS,KAUX,NAUX,NSHD,ADINIT);
  int N=0;
  for(int i=0;i<CODE->GetSize();i++)
  {
    Program *p=(Program*) CODE->At(i);
    int n=p->getNComp();
    for(int j=0;j<n;j++) 
    {
      fprintf(ou,"%08x  ",p->getComp(j));
      N++;
      if(N==10) {fprintf(ou,"\n"); N=0;}
    }
  }  
  fclose(ou);
  return;
}




















