// ****************************************************************************
// Copyright (C) 2004  Alexandre Suaide
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The Crate Controler Wiener-CC32 is not authorized for use as
// critical component in life support systems without the express
// written approval of ARW Elektronik Germany.
//
// Maintainer(s): Alexandre Suaide (suaide@if.usp.br)
//
// ****************************************************************************

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>

#include "c111at-drv.h"
#include "common/common.c"


/***************************************************************/
#ifdef KERNEL26
int __init init_module(void)
#else
static int init_module(void)
#endif
{
  int i;

  if (register_chrdev(AC_MAJOR, CC_NAME, &cc_fops))
  {
    printk(" ERROR: unable to get major %i for " CC_NAME " device\n",AC_MAJOR);
    return -EIO;
  }
	
	
  for(i=0;i<MAX_CRATES;i++)
  {
    printk("===============================\nSetting up for crate # %d \n",i);
    cc_in_use[i] = 0;
    set_crate_number(i);
    camcl(4);
    camcl(1);
    camcl(2);
    camcl(4);
    camcl(DISABLE);
    camcl(5);
  }
  printk("===============================\n");
  printk (CC_NAME ": initializing module \n");
  return 0;
}
/***************************************************************/
#ifdef KERNEL26
void __exit cleanup_module(void)
#else
static void cleanup_module(void)
#endif
{	
  if (unregister_chrdev(AC_MAJOR, CC_NAME) != 0)
    printk(CC_NAME ": cleanup_module failed\n");
  else
  {
    free_irq(CC_IRQ,NULL);
    release_region(GEO_port,0x20);
    printk(CC_NAME ": module removed\n");
  }
}
/***************************************************************/
static int cc_open(struct inode *inode, struct file *file)
{
  int istat;
  int i;

  #ifndef KERNEL26
    printk(CC_NAME": Checking region and registering ports\n");
    init_module();
    if(check_region(GEO_port,0x20))
    {
      printk(CC_NAME": Unable to register CAMAC ports\n");
      return -EIO;
    }
  #endif
  
  printk(CC_NAME": Requesting region\n");
  if(!request_region(GEO_port,0x20,"C111aT"))
  {
    printk(CC_NAME": Unable to register CAMAC ports\n");
    return -EIO;
  }
  
  printk(CC_NAME": Requesting IRQ\n");
  istat = request_irq(CC_IRQ, cc_interrupt, SA_INTERRUPT, "C111aT" ,NULL);

  if (istat)
  {
    printk(CC_NAME": can't allocate IRQ%d for DSP camac\n",CC_IRQ);
    release_region(GEO_port,0x20);
    return -EIO;
  }
  
  for(i=0;i<MAX_CRATES;i++)
  {
    printk("===============================\nSetting up for crate # %d \n",i);
    cc_in_use[i] = 0;
    set_crate_number(i);
    camcl(4);
    camcl(1);
    camcl(2);
    camcl(4);
    camcl(DISABLE);
    camcl(5);
    init_crate(i);
  }
  if(cc_in_use[0] == 1)
  {
    printk("===============================\n");
    printk("Crate number 0 set as current\n");
    set_crate_number(0);
  }
  printk("===============================\n");
  printk("CAMAC open OK \n");
  return 0;
}
/***************************************************************/
static void init_crate(int crate)
{
  if (crate>=MAX_CRATES)
  {
    return;
  }
  if(cc_in_use[crate]==1)
  {
    printk("Crate already in Use...\n");
    return;
  }
  printk("===============================\n");
  printk("Initializing crate number %d \n",crate);
  cc_in_use[crate] = 1;
  set_crate_number(crate);
  camcl(1);
  camcl(2);
  camcl(4);
  return;
}
/***************************************************************/
static void close_crate(int crate)
{
  cc_in_use[crate] = 0;
  set_crate_number(crate);
  camcl(1);
  camcl(5);
  return;
}
/***************************************************************/
static int cc_release(struct inode *inode, struct file *file)
{
  unsigned int crate = MINOR(inode->i_rdev);
  release_region(GEO_port,0x20);
  free_irq(CC_IRQ,NULL);
  close_crate(crate);
  return 0;
}
/***************************************************************/
#ifdef KERNEL26
static irqreturn_t cc_interrupt(int irq, void *dev, struct pt_regs *regs)
#else
static void  cc_interrupt(int irq, void *dev, struct pt_regs *regs)
#endif
{
  read_event_now();
  #ifdef KERNEL26
    local_irq_enable();
    return IRQ_HANDLED;
  #else
   sti();    
   return;
  #endif
}
/***************************************************************/
static int camo16(int NFA,int D)
{
  outw_p(NFA,FNA_port);
  outw_p(D,LW_W);
  return inb(R_S) & 3;
}
/***************************************************************/
static int camo24(int NFA,int D)
{
  int BH;
  BH = (int) ((D >> 16) & 255); 
  outw_p(NFA,FNA_port);
  outw_p(D,LW_W);
  return inb(R_S) & 3;
}
/***************************************************************/
static int cami16(int NFA,int *D)
{
  int lw;
  NFA = NFA |0x8000;
  outw_p(NFA,FNA_port);
  lw = (inw_p(LW_R)&0xffff);
  udelay(1);
  *D = lw;
  return (inb_p(R_S) & 3);
}
/***************************************************************/
static int cami24(int NFA, int *D)
{
  unsigned int lw,hb;
  NFA = NFA |0x8000;
  outw_p(NFA,FNA_port);
  hb = inb_p(HB_R)&0x00ff;
  hb = inb_p(HB_R)&0x00ff;
  lw = inw_p(LW_R);
  lw = lw + (hb<<16);
  *D = lw;
  udelay(1);
  return inb_p(R_S) & 3;
}
/***************************************************************/
static void camcl(int I)
{
   int bits, opt;
   unsigned char st;
   opt=(0x000f&I);
   bits=(0x0ff00&I)>>8;
   switch(opt)
   {
     case 1:
       outb(0,Crate_Z);
       printk("Z on crate \n");
       break;
     case 2:
       outb_p(0,Crate_C);
       printk("Clear on crate \n");
       break;
     case 3:
       if(I==ENABLE)printk("Enable LAM\n");
       if(I==DISABLE)printk("Disable LAM\n");
       printk("COMMAND FOR LAM bits = %x S_mod_W = %x\n",bits,S_mod_W);
       outb(bits,S_mod_W);
       break;
     case 4:
       outb_p(0,Crate_on);
       bits=0x000B;
       outb_p(bits,S_mod_W);
       printk("Crate On-line, Clear Inhibit \n");
       break;
     case 5:
       bits=0x000A;
       outb_p(bits,S_mod_W);
       outb_p(0,Crate_off);
       printk("Crate Off-line, Set Inhibit \n");
       break;
     case 6:
       st=inb_p(R_S)&0x07;
       printk("Crate Status: %x \n",st);
     default:
       break;
   }
}
