// ****************************************************************************
// Copyright (C) 2004  Alexandre Suaide
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The Crate Controler Wiener-CC32 is not authorized for use as
// critical component in life support systems without the express
// written approval of ARW Elektronik Germany.
//
// Maintainer(s): Alexandre Suaide (suaide@if.usp.br)
//
// ****************************************************************************
#ifndef __CAMAC_PROTOTYPES__
#define __CAMAC_PROTOTYPES__

  #include <linux/version.h>
  #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
    #define KERNEL26
  #endif

  #include <linux/major.h>
  #include <linux/ioport.h>
  #include <asm/io.h>
  #include <asm/uaccess.h>
  #include <asm/delay.h>

  #ifdef KERNEL26
    #include <linux/init.h>
    #include <linux/ioctl.h>
    #include <linux/poll.h>
    #include <linux/interrupt.h>
    #include <linux/sched.h>
    #include <linux/errno.h>
  #else
    #include <asm/irq.h>
  #endif

  MODULE_LICENSE("GPL");
  MODULE_AUTHOR("Alexandre Suaide");
  MODULE_DESCRIPTION("Camac Driver for SPMRoot data acquisition system");

  #include "defines.h"

	////////////////////////////////////////////////////////////////////
	// variables
	////////////////////////////////////////////////////////////////////
	
  #ifdef PCICC32
    static unsigned int nMajor               = 0;
    CC32_DESCRIPTOR *CRATES[MAX_CRATES];
    CC32_DESCRIPTOR *IRQCRATE                = NULL;
    CC32_DESCRIPTOR *CURRENT                 = NULL;
    char bin[32];
    static int         NAF                   = 0;
  #else
    struct wait_queue *cc_wait_q;
    static int         cc_in_use[MAX_CRATES];
  #endif
  #ifdef C111AT
    static int         NFA = 0;
  #endif
  #ifdef CAMEH
    static int         NN,AA,FF,CC_IO_OP;
  #endif
  

  //static volatile short unsigned int cc_buf[BUFFER_SIZE_ARRAY];
  //static volatile unsigned int       cc_inst_buf[INSTR_SIZE];
  static short unsigned int* cc_buf = 0;
  static unsigned int*       cc_inst_buf = 0;

  static int iev = 0;
  static int MAXSIZE = 512;
  static int EVENT_SIZE = 512;
  static int FLAG[MAX_BUFFER];
  static int NEXTW;
  static int NEXTR;
  static int NBUF;
  static int NBUFFULL = 0;
  static int nwords = 0;
  static int Running = 0;
  static int lost = 0;
  static int lock = 0;
  static int lockEVO = 0;
  static int CC_IO_OP = 0;
  static int USEADACPLUS = 0;
  static int IRQCOUNT = 0;
  
  static unsigned int offset_b = 0;

  // these are the registers for the adacPlus compiler
  static unsigned int REG[NREGISTERS][MAXREGISTERS];
  static unsigned int LIMIT[NREGISTERS];
  static unsigned int REG1 = 0;
  static unsigned int REG2 = 0;
  static unsigned int R1 = 0;
  static unsigned int R2 = 0;
  static unsigned int BRURADD[MAXBRUR];
	
	////////////////////////////////////////////////////////////////////
	// prototypes
	////////////////////////////////////////////////////////////////////

  static int             cc_alloc();
  static int             cc_free();
	
  static int             cc_open(struct inode*,struct file*);
  static int             cc_release(struct inode*,struct file*);
  static ssize_t         cc_read(struct file*,char*,size_t,loff_t*);
  static ssize_t         cc_write(struct file*,const char*,size_t, loff_t*);
  static int             cc_ioctl(struct inode*,struct file*,unsigned int,unsigned long);
  
  #ifdef KERNEL26
    static irqreturn_t   cc_interrupt(int, void*,struct pt_regs*);
  #else
    static void          cc_interrupt(int, void*,struct pt_regs*);
    static int           init_module(void);
    static void          cleanup_module(void);
  #endif

  static void            read_event_now();
  static int             ev_read(unsigned short int []);
  static int             ev_read_plus(unsigned short int []);
  static int             set_crate_number(int);
  static void            init_crate(int);
  static void            close_crate(int);
  
  #ifdef C111AT
    static int             camo24(int,int);
    static int             cami16(int,int*);
    static int             cami24(int,int*);
    static int             camo16(int,int);
  #endif
  #ifdef CAMEH
    static int             camo24(int ,int ,int , int);
    static int             cami16(int ,int ,int , int *);
    static int             cami24(int ,int ,int , int *);
    static int             camo16(int ,int ,int , int);
  #endif
  #ifdef PCICC32
    static int             camo16(int,int);
    static int             camo24(int,long int);
    static int             cami16(int,int*);
    static int             cami24(int,long int*);
    static void            camcl(CC32_DESCRIPTOR *,int);
  #else
    static void            camcl(int );
  #endif

  struct file_operations cc_fops =
  {
    owner: THIS_MODULE,
    read: cc_read,
    write: cc_write,
    ioctl: cc_ioctl,
    open: cc_open,
    release: cc_release
  };

#endif
