// ****************************************************************************
// Copyright (C) 2004  Alexandre Suaide
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The Crate Controler Wiener-CC32 is not authorized for use as 
// critical component in life support systems without the express 
// written approval of ARW Elektronik Germany.
//
// Maintainer(s): Alexandre Suaide (suaide@if.usp.br)
//
// ****************************************************************************

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>

#include "pcicc32.h"
#include "common/common.c"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)  
#define kfree_s(ptr, size)  kfree(ptr)   /* a workaround cause kfree_s is disapeared in V 2.4 */
#endif


/*--- FUNCTIONS -------------------------------------------------------------------------*/
#ifdef KERNEL26
int __init init_module(void)
#else
static int init_module(void)
#endif
{
  int             i = 0;
  int             result = 0;
  PCIConfigHeader *ch;
  CC32_DESCRIPTOR *wd;
  for(i = 0;i<MAX_CRATES;i++) CRATES[i] = NULL;
  
  i = 0;
  printk("\n========================================================\n");
  printk(  "==                                                    ==\n");
  printk(  "== pcicc32 : init_module                              ==\n");
  printk(  "==                                                    ==\n");
  printk(  "========================================================\n");
  
  /* search for all PCIADA modules */
  while ((ch = GetPCIConfigHeader(ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, i)))
  {
    printk("---------------------------------------\n");
    printk("Subsystem Id = %d\n",ch->subsystem_id);
    printk("Subsystem vendor = %d\n",ch->subsystem_vendor_id);
    printk("desc[3].size = %d\n",ch->desc[3].size );
    if(ch->subsystem_id != ARWVME_SUBSYSTEM_ID &&
       ch->subsystem_vendor_id != ARWVME_SUBSYSTEM_VENDOR_ID &&
       ch->desc[3].size != 0x8000) kfree_s(ch, sizeof(*ch));        // FREE(ch);
    else
		{
      printk("pcicc32 : found %d. PCIADA card\n", i + 1);
      printk("---------------------------------------\n");
  	  PLX9050BugFix(ch);
  	  wd = (CC32_DESCRIPTOR *)kmalloc(sizeof(CC32_DESCRIPTOR), GFP_ATOMIC);
	    i++;
	        
			init_waitqueue_head(&wd->event_queue);    
		  wd->pLCR = wd->pUsr = 0;
	    wd->pPch = ch;
		  wd->cModuleNumber = 255;
		  wd->cFPGAVersion = 255;
		  wd->bConnected = 0;
		  wd->wInitStep = 0;     
	    wd->wIrq = 0xFFFF;          
	    wd->dwInterruptCount = 0;
	    wd->wIrqStatus = 0;  
		  wd->wIndex = i;			
			
	    if(!request_io_memory(ch)) // not means ok
	    {
	      wd->wInitStep = 1;	    
		    if(translate_addresses(wd, ch)) 
        { 
          printk("pcicc32 : translation of addresses failed!\n"); 
          kfree_s(wd,sizeof(*wd)); 
        }
		    else // successful translate_addresses
		    {
		      wd->wInitStep = 2;		    
			    wd->bConnected =  get_module_info(wd, &wd->cModuleNumber, &wd->cFPGAVersion);
			    if (wd->bConnected && test_connection(wd))
			    { 
			      printk("pcicc32 : connection test failed!\n");
				    wd->bConnected = 0;
			    }
				  else if(wd->cModuleNumber<MAX_CRATES) CRATES[wd->cModuleNumber] = wd; 
			  }
	    }
	    else
	    {
		    printk("pcicc32 : requested io-memory still claimed!\n");
		   	kfree_s(wd, sizeof(*wd));       // FREE(wd);
		  }
		}	
  }
	
	// requesting IRQ for the smallest crate ID
	for(i=0;i<MAX_CRATES;i++) if(!IRQCRATE && CRATES[i])
	{
	  wd = CRATES[i];
		if(request_irq(wd->pPch->irq, cc_interrupt, SA_INTERRUPT | SA_SHIRQ, DEVICE_NAME, wd))
	  {
		  printk("pcicc32 : can't get irq @ %d\n", wd->pPch->irq);
			kfree_s(wd, sizeof(*wd));       // FREE(wd);
			IRQCRATE = NULL;
			CRATES[i] = NULL;
	  }
	  else
	  {   
	  	// successful request_irq
			printk("pcicc32 : IRQ requested ok IRQ = %d to crateId = %d Crate = 0x%x\n",
			wd->pPch->irq,wd->cModuleNumber,(u32)wd);
		  wd->wInitStep = 3;
		  wd->wIrq = wd->pPch->irq;
			IRQCRATE = wd;		
		}
  }
  
  nMajor = MAJOR_NO;
  result = register_chrdev(nMajor, DEVICE_NAME, &cc_fops);
  if (result < 0)
  {
  	printk("pcicc32: Can't install driver (%d)\n", result);	
		cleanup_module();
  }
  else printk("pcicc32 : major #%d assigned.\n", nMajor);  
  /* register the proc device */   
  return create_proc_read_entry(DEVICE_NAME, 0, NULL, pcicc32_read_proc, NULL) ? 0 : -ENODEV; 
}
static int request_io_memory(PCIConfigHeader *pPch)
{
	if (check_mem_region(pPch->desc[0].base_address, pPch->desc[0].size)) return -EBUSY;	
	if (check_mem_region(pPch->desc[3].base_address, pPch->desc[3].size)) return -EBUSY;	
	request_mem_region(pPch->desc[0].base_address, pPch->desc[0].size, DEVICE_NAME);
	request_mem_region(pPch->desc[3].base_address, pPch->desc[3].size, DEVICE_NAME);
	return 0;
}
static void release_io_memory(PCIConfigHeader *pPch)
{
	release_mem_region(pPch->desc[0].base_address, pPch->desc[0].size);
	release_mem_region(pPch->desc[3].base_address, pPch->desc[3].size);
}
static int translate_addresses(CC32_DESCRIPTOR *pWd, PCIConfigHeader *pPch)    /* differs from PCIVME */
{
	if (pPch->desc[0].type == PCI_BASE_ADDRESS_MEM_TYPE_1M) pWd->pLCR = (u32)bus_to_virt(pPch->desc[0].base_address);  /* LCR ISA base addresses */
	else pWd->pLCR = (u32)ioremap(pPch->desc[0].base_address, pPch->desc[0].size);

	if (pPch->desc[3].type == PCI_BASE_ADDRESS_MEM_TYPE_1M) pWd->pUsr = (u32)bus_to_virt(pPch->desc[3].base_address); /* User ISA base addresses */
	else pWd->pUsr = (u32)ioremap(pPch->desc[3].base_address, pPch->desc[3].size);	
	return 0;
}
static void un_translate_addresses(CC32_DESCRIPTOR *pWd, PCIConfigHeader *pPch)
{
	if (pPch->desc[0].type != PCI_BASE_ADDRESS_MEM_TYPE_1M) iounmap((void *)pWd->pLCR);
	if (pPch->desc[3].type != PCI_BASE_ADDRESS_MEM_TYPE_1M) iounmap((void *)pWd->pUsr);
}
int test_connection(CC32_DESCRIPTOR *wd)
{
	u16 intCSR_store;
	u16 cntrl_store;
	int i = 1000;
	u32 access_store;
	int error = 0;
	
	cntrl_store  = readl(wd->pLCR + PLX9050_CNTRL);  /* read CONTROL register */
	intCSR_store = readl(wd->pLCR + PLX9050_INTCSR); /* read interrupt + CSR register */
	
	writel(0, wd->pLCR + PLX9050_INTCSR); /* disable interrupts */
	writew(cntrl_store | 0x0180, wd->pLCR + PLX9050_CNTRL); /* enable access */
	
	access_store = readl(wd->pUsr + MAKE_CC32_ADR(26,1,0));
	while (i--)
	{
		writel(0x55555555, wd->pUsr + MAKE_CC32_ADR(26,1,0));
		if (0x00555555 != readl(wd->pUsr + MAKE_CC32_ADR(26,1,0))) { error = 1; break; }
		writel(0xAAAAAAAA, wd->pUsr + MAKE_CC32_ADR(26,1,0));
		if (0x00AAAAAA != readl(wd->pUsr + MAKE_CC32_ADR(26,1,0))) { error = 1; break; }
		writel(0xFFFFFFFF, wd->pUsr + MAKE_CC32_ADR(26,1,0));
		if (0x00FFFFFF != readl(wd->pUsr + MAKE_CC32_ADR(26,1,0))) { error = 1; break; }
		writel(0x00000000, wd->pUsr + MAKE_CC32_ADR(26,1,0));
		if (0x00000000 != readl(wd->pUsr + MAKE_CC32_ADR(26,1,0))) { error = 1; break; }
	}	
	writew(cntrl_store & ~0x0100, wd->pLCR + PLX9050_CNTRL); /* clear potential interrupt */	
	/* restore all contents */
	writel(access_store, wd->pUsr + MAKE_CC32_ADR(26,1,0));
	writew(cntrl_store, wd->pLCR + PLX9050_CNTRL);
	writew(intCSR_store, wd->pLCR + PLX9050_INTCSR);	
	return error; 	
}
int get_module_info(CC32_DESCRIPTOR *wd, u8 *cModuleNumber, u8 *cFPGAVersion)
{
	u16 intCSR_store;
	u16 cntrl_store;
	int found = 0;
	u16 data;
	
	cntrl_store  = readl(wd->pLCR + PLX9050_CNTRL);  /* read CONTROL register */
	intCSR_store = readl(wd->pLCR + PLX9050_INTCSR); /* read interrupt + CSR register */
	
	printk("pcicc32 : cntrl=0x%04x, intCSR=0x%04x\n", cntrl_store, intCSR_store); 
	
	if (cntrl_store & 0x0800) /* a CC32 is connected */
	{
		u16 bla = cntrl_store | 0x0180;		
		writew(0, wd->pLCR + PLX9050_INTCSR); /* disable interrupts */
		writew(bla, wd->pLCR + PLX9050_CNTRL); /* enable access */		
		data = readl((wd->pUsr + MAKE_CC32_ADR(0,0,0)));
		printk("pcicc32 : CC32 status=0x%04x\n", data); 
		if ((data & 0xF000) != 0x8000)
		{
			*cModuleNumber = *cFPGAVersion = 255;
			printk("pcicc32 : Wrong module type connected @ index %d (0x%04x)!\n", wd->wIndex, data);
		}
		else
		{
			found = 1;
			*cModuleNumber = (data >> 4) & 0xF;
			*cFPGAVersion  = (data >> 8) & 0xF; 
		} 
		writew(cntrl_store & ~0x0080, wd->pLCR + PLX9050_CNTRL); /* clear potential interrupt */	
		/* restore all contents */
		writew(cntrl_store, wd->pLCR + PLX9050_CNTRL);
		writew(intCSR_store, wd->pLCR + PLX9050_INTCSR);
	}	
	return found; 	
}
static int pcicc32_read_proc(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
  int i=0;
  int pos = 0;
  CC32_DESCRIPTOR *pd;
  PCIConfigHeader *ch;
  u16 		        cntrl;
  char            *cause = "None";
  
  pos += sprintf(buf + pos, "\npcicc32 information. Version %d.%d of %s.\n", VERSION_HI, VERSION_LO, __DATE__);  
  pos += sprintf(buf + pos, " ---------------------\n"); 
  pos += sprintf(buf + pos, " Interfaces found : %d\n", i);
  
  for(i=0;i<MAX_CRATES;i++) if(CRATES[i])
  {
  	pd = (CC32_DESCRIPTOR *)CRATES[i];
  	ch = pd->pPch;
	  cntrl = readl(pd->pLCR + PLX9050_CNTRL);
    pos += sprintf(buf + pos, " --- %d ---------------\n", i + 1); 
	  pos += sprintf(buf + pos, " LCR  phys/virt/size : 0x%08x/0x%08x/%d\n", ch->desc[0].base_address, pd->pLCR, ch->desc[0].size);
	  pos += sprintf(buf + pos, " User phys/virt/size : 0x%08x/0x%08x/%d\n", ch->desc[3].base_address, pd->pUsr, ch->desc[3].size);
	  pos += sprintf(buf + pos, " Irq                 : %d\n", pd->wIrq);
	  
	  if (pd->bConnected)
	  {
		  pos += sprintf(buf + pos, " CC32 is or was      : (software) connected.\n");
	  	pos += sprintf(buf + pos, " Module-Number       : %d\n", pd->cModuleNumber);
	  	pos += sprintf(buf + pos, " FPGA-Version        : %d\n", pd->cFPGAVersion);
		}
		else pos += sprintf(buf + pos, " CC32 is or was      : not (software) connected.\n");		
		
	  if (!((cntrl & 0x0800) && (!(cntrl & 0x0600)))) pos += sprintf(buf + pos, " CC32 is             : powered off or cable disconnected.\n");
		
	  pos += sprintf(buf + pos, " IrqCount            : %d\n", pd->dwInterruptCount);
	  
    if (pd->wIrqStatus & PCIADA_INTERRUPT) cause = "Timeout";
	  else if (pd->wIrqStatus & CC32_INTERRUPT) cause = "LAM";	  
	  
    pos += sprintf(buf + pos, " Pending IrqStatus   : %s\n", cause);	  
  }

  pos += sprintf(buf + pos, "\n");
  
  *eof = 1;
  
  return pos;
}
#ifdef KERNEL26
static irqreturn_t cc_interrupt(int irq, void *dev, struct pt_regs *regs)
#else
static void  cc_interrupt(int irq, void *dev, struct pt_regs *regs)
#endif
{
  int T = 0;
	//cli();
  if (IRQCRATE)
  {
    // evaluate the reason of the interrupt - if it is mine
	  T = my_interrupt(readw(IRQCRATE->pLCR + PLX9050_INTCSR));
		if (T == CC32_INTERRUPT)
	  {
	    //writew(DISABLE_PCIADA_IRQS, IRQCRATE->pLCR + PLX9050_INTCSR);       /* disable global interrupts */
	    IRQCRATE->wIrqStatus = CC32_INTERRUPT;
	    IRQCRATE->dwInterruptCount++;
		  //writel(0x00000000 ,IRQCRATE->pUsr + MAKE_CC32_ADR(28,0,16)); // clear LAM mask register
			read_event_now();
			writel(0x00000000 ,IRQCRATE->pUsr + MAKE_CC32_ADR(28,0,16)); // clear LAM mask register
	    //writew(ENABLE_PCIADA_IRQS, IRQCRATE->pLCR + PLX9050_INTCSR);       /* enable global interrupts */
	    //writew(ENABLE_ACCESS_NOAUTOREAD, IRQCRATE->pLCR + PLX9050_CNTRL);       /* enable global interrupts */
		}  
  }
  //sti();
	return;
}
static int my_interrupt(u16 intCSR)
{
  int result = NOT_MY_INTERRUPT;  
  if (intCSR & 0x0040)  // it is global enabled
  {    
    if ((intCSR & 0x0005) == 0x0005) { result = CC32_INTERRUPT; return result;}  // it is a enabled CC32 interrupt         
    if ((intCSR & 0x0028) == 0x0028) { result = PCIADA_INTERRUPT; return result;} // it is a enabled PCIADA interrupt      
  }
  return result;
}
#ifdef KERNEL26
void __exit cleanup_module(void)
#else
static void cleanup_module(void)
#endif
{
  printk("Cleaning pcicc32 module\n");
  int i;
  PCIConfigHeader *ch;
  CC32_DESCRIPTOR *wd;
  unregister_chrdev(nMajor, DEVICE_NAME);  
  /* unregister the proc device */
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,19)
  proc_unregister(&proc_root, pcimod_proc_entry.low_ino);
  #else
  remove_proc_entry(DEVICE_NAME, NULL);
  #endif    
  for(i=0;i<MAX_CRATES;i++) if(CRATES[i])
  {
	  wd  = (CC32_DESCRIPTOR *)CRATES[i];
  	ch = wd->pPch;
	  switch (wd->wInitStep)
	  {
	    case 3:  writew(readl(wd->pLCR + PLX9050_INTCSR) & ~0x40, wd->pLCR + PLX9050_INTCSR);  // disable global interrupts
               free_irq(wd->wIrq, wd);  
	    case 2:  un_translate_addresses(wd, ch);
	    case 1:  release_io_memory(ch);
	    default: wd->wInitStep = 0; 
	  }
  }  
  for(i=0;i<MAX_CRATES;i++) CRATES[i] = NULL;
  IRQCRATE      = NULL;
  CURRENT       = NULL;
  return;
}
static void switch_CC32_on(CC32_DESCRIPTOR *desc)
{
	printk("TURNING CRATE %d ON...\n",desc->cModuleNumber);

	writew(ENABLE_ACCESS_NOAUTOREAD, desc->pLCR + PLX9050_CNTRL); /* enable access */
  writew(DISABLE_PCIADA_IRQS, desc->pLCR + PLX9050_INTCSR);     /* disable global interrupts */
	camcl(desc,0);
	camcl(desc,5);
	camcl(desc,2);
	camcl(desc,1);

}
static void switch_CC32_off(CC32_DESCRIPTOR *desc)
{	
 	printk("TURNING CRATE %d OFF...\n",desc->cModuleNumber);
  writew(DISABLE_PCIADA_IRQS, desc->pLCR + PLX9050_INTCSR);     /* disable global interrupts */
	writew(DISABLE_ACCESS, desc->pLCR + PLX9050_CNTRL);           /* disable access */
}
// this method enables the connection to ALL the available crates
// but the struct file *pFile returns only the pointer to the
// first module found
int cc_open(struct inode *pInode, struct file *pFile)
{
	cleanup_module();
  init_module();
  CC32_DESCRIPTOR *wd = 0;
	int i,status = 0;
	int nMinor = MINOR(pInode->i_rdev);	
	int nMajor = MAJOR(pInode->i_rdev);
 	printk("pcicc32 : open(), Major=%d, Minor=%d\n", nMajor, nMinor);
 
  for(i = 0;i<MAX_CRATES;i++)
  {
    if(CRATES[i])
    {
      wd  = CRATES[i];
      wd->bConnected =  get_module_info(wd, &wd->cModuleNumber, &wd->cFPGAVersion);
		  if(wd->bConnected) 
      {
        if(test_connection(wd))
        {
				  printk("pcicc32 : connection test for module %d failed!\n", wd->cModuleNumber);
				  wd->bConnected = 0;      
        }
				else 
        {
				  printk("pcicc32 : connection test for module %d is OK!\n", wd->cModuleNumber);
          printk("Status of crate selection %d\n",set_crate_number(i));
          printk("pcicc32 : found CC32 module with number %d.\n", wd->cModuleNumber);
 	        switch_CC32_on(wd);			
					status = 1;
       }
      }
      else printk("pcicc32 : module %d not connected!\n", nMinor);	
    }
  }
	if (status==1)
	{
		pFile->private_data = (void *)kmalloc(sizeof(CC32_DESCRIPTOR), GFP_ATOMIC);
		if (!pFile->private_data) return -ENOMEM;		
	}
	else
	{
		printk("pcicc32 : No CC32 module found.\n");
		return -ENODEV;
	}				
	MOD_INC_USE_COUNT;
	return 0;
}
// this method releases the connection to ALL the available crates
int cc_release(struct inode *pInode, struct file *pFile)
{
	int i;
  for(i=0;i<MAX_CRATES;i++)
  {
    if(CRATES[i]) switch_CC32_off(CRATES[i]);
  }
  void *pd = pFile->private_data;
  if(pd)
  {
		kfree_s(pd, sizeof(CC32_DESCRIPTOR)); 
    pFile->private_data = 0;
  } 
	MOD_DEC_USE_COUNT;
	return 0;
}
/*
------------------------------------------------------------------------
------------------------------------------------------------------------
------------------------------------------------------------------------
------------------------------------------------------------------------
Now it starts the Pelletron specific stuf
------------------------------------------------------------------------
------------------------------------------------------------------------
------------------------------------------------------------------------
------------------------------------------------------------------------
*/
/***************************************************************/
static int camo16(int NAF,int D)
{
  writew(D,CURRENT->pUsr+NAF);
  return 0;
}
/***************************************************************/
static int camo24(int NAF,long int D)
{
  writew(D,CURRENT->pUsr+NAF);
  return 0;
}
/***************************************************************/
static int cami16(int NAF,int *D)
{
  *D = readl(CURRENT->pUsr+NAF);
  return 0;
}
/***************************************************************/
static int cami24(int NAF, long int *D)
{  
  *D = readl(CURRENT->pUsr+NAF);
  return 0;
}
/***************************************************************/
static void camcl(CC32_DESCRIPTOR *wd,int c)
{
  unsigned short int dummy[MAX_EVENT_SIZE];
  int i = 0;
	switch(c)
	{
		case 0: // RESET on crate
		  printk("RESET ON CRATE\n");
			writew(1 ,wd->pUsr + MAKE_CC32_ADR(31,0,16));
		  break;
		case 1: // Z on crate
		  printk("Z ON CRATE\n");
			writew(1 ,wd->pUsr + MAKE_CC32_ADR(0,1,16));
		  break;
		case 2: // C on crate
		  printk("C ON CRATE\n");
		  writew(1 ,wd->pUsr + MAKE_CC32_ADR(0,0,16));
		  break;
		case 3: // LAM enable and RUN
			printk("\n ----------- ENABLE IRQ and RUN\n");
			printk("ENABLE IRQ = %d to crateId = %d Crate = 0x%x\n", wd->pPch->irq,wd->cModuleNumber,(u32)wd);
		  camcl(wd,5);									
		  writel(0x00FFFFFF ,wd->pUsr + MAKE_CC32_ADR(28,1,16)); // LAM mask register			
			for(i=0;i<24;i++) readw(wd->pUsr + MAKE_CC32_ADR(i,0,10)); // clear all LAM
			writel(0x00000000 ,wd->pUsr + MAKE_CC32_ADR(28,0,16)); // LAM mask register
      
			if(USEADACPLUS==0) ev_read(dummy);
      else ev_read_plus(dummy);
			writew(ENABLE_PCIADA_IRQS ,wd->pLCR + PLX9050_INTCSR);
	    		  
			break;
		case 4: // LAM disable and STOP
		  printk("\n ----------- DISABLE IRQ and STOP\n");
			printk("DISABLE IRQ = %d to crateId = %d Crate = 0x%x\n", wd->pPch->irq,wd->cModuleNumber,(u32)wd);
		  camcl(wd,6);
	    writew(DISABLE_PCIADA_IRQS ,wd->pLCR + PLX9050_INTCSR);
	    break;
		case 5: // Clear Inhibit
		  printk("Clear inhibit\n");
			writew(1 ,wd->pUsr + MAKE_CC32_ADR(27,1,16));
		  break;
		case 6: // Set Inhibit
		  printk("Set inhibit\n");
			writew(1 ,wd->pUsr + MAKE_CC32_ADR(27,0,16));
		  break;
		default:
		  break;
	}
	return;
}
char* tobin(u32 input)
{
	u32 tmp = input;
	u32 r = 0;
	u32 bit = 32;
	u32 i;
	for(i=0;i<32;i++) bin[i] = '0';
	do
	{
	  r = tmp%2;
		tmp/=2;
		bit--;
		if(r==1) bin[bit] = '1';
  } while (tmp>0);
	return &bin[0];
}
void printNAF(CC32_DESCRIPTOR *wd,int N,int A,int F)
{
	u32 d = readl(wd->pUsr + MAKE_CC32_ADR(N,A,F));
	printk("N = %2d | A = %2d | F = %2d = 0x%08x - %s\n",N,A,F,d,tobin(d));
}

