#include "TH2.h"
#include "ScanRoot.h"

TH1F* h1[100];
TH2F* h2[100];

ScanRoot* myScan;

/////////////////////////////////////////////////////////////////////
void fillHistograms(short npar, float *par)
{
  if(npar==0) return;
  if(par[2]>0 && par[1]>0) h2[0]->Fill(par[2],par[1]);
  
  // create a new parameter to be included in a
  // .FIL file.
  //
  // the parameter will be added to the .FIL file
  // if the user is saving the .FIL in the SPMRoot
  // or if there is an output file open in ScanRoot
  // the function ScanRoot::addPar(int,float) adds
  // a new parameter to the output stream. This new
  // parameter is not availabe in the par[] vector
  // right away but only in the new .FIL file. 
  // the new parameter will be availabe in the par[]
  // vector only when the new .FIL file is read again
  //
  // the parameter can be added either using the software
  // trigger or the histogramming function. It does not matter
  // unless you want to remove events from the file that does
  // not have the new parameter.
  //
  // it is important to remind that the .FIL format requires
  // that any parameter should be converted to a 15 bits
  // integer (0 to 32767). The parameter in the function is
  // a flota but the program truncates the value to an
  // integer in that range
  //
  float parValue = par[1]+par[2];
  int parNumber = 100;  
  if(parValue>0) myScan->addPar(parNumber,parValue);

}
/////////////////////////////////////////////////////////////////////
void bookHistograms(ScanRoot* scan)
{
  // this next line is to be sure that the histograms are
  // saved in the correct ROOT directory
  scan->cd();
  
  // creating histograms
  h2[0] = new TH2F("Det_1","Det_1",512,0,4096,512,0,4096);
  
  // this line links the histogram filling function
  // to the acquisition program.
  scan->fillHistograms = &fillHistograms;
  
  // need to save the pointer position in a global
  // variable because we will use that in another
  // function. In this example, we will use this
  // pointer to add new parameters to the .FIL
  // file while processing the events
  myScan = scan;
    
  return;
}
/////////////////////////////////////////////////////////////////////
void init()
{
}
/////////////////////////////////////////////////////////////////////
void finish()
{
}
